package hexagon.graphics;

import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL20.*;
import hexagon.Main;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.function.Consumer;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL20;
import org.lwjgl.util.vector.Matrix4f;

public class Shader {
	
	private static final ArrayList<Shader> shaders = new ArrayList<Shader>();
	
	public static Shader hexagon, gui;
	
	static {
		hexagon = new Shader("hexagon", new String[] { "positions", "texCoords" });
		gui = new Shader("gui", new String[] { "position" });
	}
	
	public static void init() {}
	
	private static FloatBuffer matrixBuffer = BufferUtils.createFloatBuffer(16);
	
	private HashMap<String, Integer> uniformLocations = new HashMap<String, Integer>();
	private HashMap<String, Integer> attributeLocations = new HashMap<String, Integer>();
	
	public final int shaderID;
	
	public Shader(String name, String... attributeBinds) {
		this("res/shaders/" + name + ".vert", "res/shaders/" + name + ".frag", attributeBinds);
	}
	
	public Shader(String vertPath, String fragPath, String... attributeBinds) {
		shaderID = glCreateProgram();
		load(vertPath, fragPath, attributeBinds);
		shaders.add(this);
	}
	
	public int getUniformLocation(String name) {
		Integer loc = uniformLocations.get(name);
		if (loc == null) {
			loc = glGetUniformLocation(shaderID, name);
			uniformLocations.put(name, loc);
		}
		return loc;
	}
	
	public int getAttribLocation(String name) {
		Integer loc = attributeLocations.get(name);
		if (loc == null) {
			loc = glGetAttribLocation(shaderID, name);
			attributeLocations.put(name, loc);
		}
		return loc;
	}
	
	public void bindAttribute(int loc, String name) {
		GL20.glBindAttribLocation(shaderID, loc, name);
	}
	
	public void setUniform1f(String name, float a) {
		glUniform1f(getUniformLocation(name), a);
	}
	
	public void setUniform2f(String name, float a, float b) {
		glUniform2f(getUniformLocation(name), a, b);
	}
	
	public void setUniform3f(String name, float a, float b, float c) {
		glUniform3f(getUniformLocation(name), a, b, c);
	}
	
	public void setUniform4f(String name, float a, float b, float c, float d) {
		glUniform4f(getUniformLocation(name), a, b, c, d);
	}
	
	public void setUniformMat4(String name, Matrix4f matrix) {
		matrix.store(matrixBuffer);
		matrixBuffer.flip();
		glUniformMatrix4(getUniformLocation(name), false, matrixBuffer);
		matrixBuffer.clear(); // maybe unnecessary
	}
	
	public void enable() {
		glUseProgram(shaderID);
	}
	
	public void delete() {
		glDeleteProgram(shaderID);
	}
	
	public void finalize() throws Throwable {
		delete();
		super.finalize();
	}
	
	public static void forEachShader(Consumer<Shader> action) {
		shaders.forEach(action);
	}
	
	private void load(String vertPath, String fragPath, String... attributeBinds) {
		int vertexShader = createShader(GL_VERTEX_SHADER, vertPath);
		int fragmentShader = createShader(GL_FRAGMENT_SHADER, fragPath);
		
		glAttachShader(shaderID, vertexShader);
		glAttachShader(shaderID, fragmentShader);
		
		int index = 0;
		for (String s : attributeBinds)
			bindAttribute(index++, s);
		
		glLinkProgram(shaderID);
		glValidateProgram(shaderID);
		
		glDeleteShader(vertexShader);
		glDeleteShader(fragmentShader);
	}
	
	private int createShader(int type, String path) {
		int shaderID = glCreateShader(type);
		StringBuilder source = new StringBuilder();
		
		try {
			BufferedReader reader = new BufferedReader(new FileReader(path));
			String line;
			while ((line = reader.readLine()) != null)
				source.append(line).append('\n');
			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
			Main.destroy();
			System.exit(1);
		}
		
		glShaderSource(shaderID, source);
		glCompileShader(shaderID);
		if (glGetShaderi(shaderID, GL_COMPILE_STATUS) == GL_FALSE) System.err.println((type == GL_VERTEX_SHADER ? "Vertexshader" : type == GL_FRAGMENT_SHADER ? "Fragmentshader" : "Shader") + " failed to compile.\nError log:\n" + glGetShaderInfoLog(shaderID, 1024));
		return shaderID;
	}
	
}
