package hexagon.graphics;

import java.nio.ByteBuffer;
import java.nio.ShortBuffer;
import java.util.function.BiConsumer;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;
import org.lwjgl.util.vector.Vector4f;

import hexagon.Main;
import hexagon.model.HexagonModel;
import hexagon.model.Loader;
import hexagon.world.BlockSide;
import hexagon.world.entity.Hexagon;

public class BlockRenderer<T> {
	
	private static Vector4f[] hexPos4f = new Vector4f[6];
	
	public static final BlockRenderer<Hexagon> hexagon;
	public static final BlockRenderer<BlockSide> blockSide;
	
	static {
		hexagon = new BlockRenderer<>(3 * 4 + 2 * 4 + 3 * 4, 6, 256, (h, renderer) -> {
			for (int i = 0; i < renderer.verticesPerObject; i++) {
				renderer.buffer.putFloat(h.block.blockColumn.points[i].x * h.globalScale + Main.main.camera.position.getEucl().x * (1f - h.globalScale));
				renderer.buffer.putFloat(h.block.blockColumn.points[i].y * h.globalScale);
				renderer.buffer.putFloat(h.block.blockColumn.points[i].z * h.globalScale);
				renderer.buffer.putFloat(Hexagon.points[i].x);
				renderer.buffer.putFloat(Hexagon.points[i].y);
				renderer.buffer.putFloat(h.colorTint.x);
				renderer.buffer.putFloat(h.colorTint.y);
				renderer.buffer.putFloat(h.colorTint.z);
			}
		});
		blockSide = new BlockRenderer<>(3 * 4 + 2 * 4 + 3 * 4, 4, 256, (s, renderer) -> {
			for (int i = 0; i < renderer.verticesPerObject; i++) {
				int x = (s.side + i / 2) % 6, y = (i + 1) % 4 / 2;
				Hexagon hex = y == 0 ? s.block.up : s.block.down;
				float mult = hex.globalScale;
				renderer.buffer.putFloat(s.block.blockColumn.points[x].x * mult + Main.main.camera.position.getEucl().x * (1f - mult));
				renderer.buffer.putFloat(s.block.blockColumn.points[x].y * mult);
				renderer.buffer.putFloat(s.block.blockColumn.points[x].z * mult);
				renderer.buffer.putFloat((x + 6 - s.side) % 6);
				renderer.buffer.putFloat(y);
				renderer.buffer.putFloat(hex.colorTint.x);
				renderer.buffer.putFloat(hex.colorTint.y);
				renderer.buffer.putFloat(hex.colorTint.z);
			}
		});
	}
	
	
	private final int vertexSizeInBytes;
	private final int objectSizeInBytes;
	private final int verticesPerObject;
	private final int numberOfObjects;
	private final int bufferSize;
	private final int numIndices;
	private final int vaoID, vboID, iboID;
	private final BiConsumer<T, BlockRenderer<T>> sumbitFunc;
	
	private ByteBuffer buffer;
	private int indicesUsed = 0;
	
	static {
		for (int i = 0; i < 6; i++)
			hexPos4f[i] = new Vector4f(HexagonModel.vertices[i * 3], HexagonModel.vertices[i * 3 + 1], HexagonModel.vertices[i * 3 + 2], 1);
	}
	
	
	public BlockRenderer(int vertexSizeInBytes, int verticesPerObject, int numberOfObjects, BiConsumer<T, BlockRenderer<T>> sumbitFunc) {
		this.vertexSizeInBytes = vertexSizeInBytes;
		this.verticesPerObject = verticesPerObject;
		this.objectSizeInBytes = verticesPerObject * vertexSizeInBytes;
		this.numberOfObjects = numberOfObjects;
		this.bufferSize = numberOfObjects * this.objectSizeInBytes;
		this.numIndices = numberOfObjects * (verticesPerObject - 2) * 3;
		this.sumbitFunc = sumbitFunc;
		vaoID = GL30.glGenVertexArrays();
		vboID = GL15.glGenBuffers();
		iboID = GL15.glGenBuffers();
		Loader.theLoader.vaos.add(vaoID);
		Loader.theLoader.vbos.add(vboID);
		Loader.theLoader.vbos.add(iboID);
		
		GL30.glBindVertexArray(vaoID);
		buffer = BufferUtils.createByteBuffer(bufferSize);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboID);
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, bufferSize, GL15.GL_STREAM_DRAW);
		GL20.glEnableVertexAttribArray(0);
		GL20.glEnableVertexAttribArray(1);
		GL20.glEnableVertexAttribArray(2);
		GL20.glVertexAttribPointer(0, 3, GL11.GL_FLOAT, false, vertexSizeInBytes, 0); // position
		GL20.glVertexAttribPointer(1, 2, GL11.GL_FLOAT, false, vertexSizeInBytes, 3 * 4); // texCoords
		GL20.glVertexAttribPointer(2, 3, GL11.GL_FLOAT, false, vertexSizeInBytes, 5 * 4); // texCoords
		
		ShortBuffer indices = BufferUtils.createShortBuffer(numIndices);
		for (short i = 0; i < numberOfObjects; i++) {
			for (short j = 0; j < verticesPerObject - 2; j++) {
				indices.put((short) (verticesPerObject * i));
				indices.put((short) (verticesPerObject * i + j + 1));
				indices.put((short) (verticesPerObject * i + j + 2));
			}
		}
		indices.flip();
		
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, iboID);
		GL15.glBufferData(GL15.GL_ELEMENT_ARRAY_BUFFER, indices, GL15.GL_STATIC_DRAW);
	}
	
	/**	Usage: begin, submit ..., end, render */
	public void submit(T t) {
		if (buffer.remaining() < objectSizeInBytes) flush();
		
		sumbitFunc.accept(t, this);
	}
	
	public void flush() {
		end();
		render();
		begin();
	}
	
	public void begin() {
		buffer.clear();
	}
	
	public void end() {
		indicesUsed = buffer.position() * 12 / objectSizeInBytes;
		buffer.flip();
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboID);
		GL15.glBufferSubData(GL15.GL_ARRAY_BUFFER, 0, buffer);
	}
	
	public void render() {
		GL30.glBindVertexArray(vaoID);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, iboID);
		GL13.glActiveTexture(GL13.GL_TEXTURE0);
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, Hexagon.model.getTextureID());
		
		GL11.glDrawElements(GL11.GL_TRIANGLES, indicesUsed, GL11.GL_UNSIGNED_SHORT, 0);
	}
	
}
