package hexagon.world;

import java.awt.Point;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.function.BiFunction;
import java.util.function.Consumer;

import org.lwjgl.input.Mouse;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;

import hexagon.Camera;
import hexagon.Main;
import hexagon.graphics.BlockRenderer;
import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.util.CylVec3;
import hexagon.util.MathUtil;
import hexagon.util.Vec3f;
import hexagon.world.entity.Hexagon;

public class World {
	
	public static final float viewDistance = 5 * (HexagonModel.y60 * 2); // Change first number only!
	public static final double angleHalfHexagon = Math.PI / (Chunk.ringSize * Chunk.chunkSize);
	public static final double radius = HexagonModel.y60 / angleHalfHexagon;
	
	private Hexagon hoverHexagon;
	private HashMap<Long, Chunk> chunks = new HashMap<>();
	private HashSet<Point> loadedChunks = new HashSet<>();
	
	private BiFunction<Integer, Integer, Consumer<Chunk>> tempBlockGenerator = (x, z) -> c -> {
		BlockColumn bc = null;
		for (int i = 0; i < Chunk.chunkSize; i++) {
			for (int j = 0; j < Chunk.chunkSize; j++) {
				bc = c.blockCols[i + j * Chunk.chunkSize] = new BlockColumn(i, j, c);
				if (Math.random() < 0.8) bc.setBlock(new Block(x * Chunk.chunkSize + i, 0, z * Chunk.chunkSize + j, bc), 0);
				if (Math.random() < 0.9) bc.setBlock(new Block(x * Chunk.chunkSize + i, 1, z * Chunk.chunkSize + j, bc), 1);
				bc.setBlock(new Block(x * Chunk.chunkSize + i, 2, z * Chunk.chunkSize + j, bc), 2);
			}
		}
	};
	
	public World() {
	}
	
	public void tick(Camera camera) {
		updateLoadedChunks(camera);
		
		updateHoverHexagon();
	}
	
	private void updateLoadedChunks(Camera camera) {
		Point chunkToLoad = null, p = new Point(), c;
		Vector2f cVec = new Vector2f(), rVec = new Vector2f(); // chunk, real
		rVec.x = camera.position.getCyl().x / Chunk.chunkSize;
		rVec.y = (float) (camera.position.getCyl().rv / camera.position.getCyl().r * radius / Chunk.chunkSize);
		float camX = rVec.x;
		float camZ = rVec.y;
		
		if (loadedChunks.isEmpty()) {
			MathUtil.realCoordsToChunk(rVec, cVec);
			chunkToLoad = new Point((int) cVec.x, (int) cVec.y);
		}
		
		for (Iterator<Point> it = loadedChunks.iterator(); it.hasNext();) {
			c = it.next();
			int cx = c.x, cz = c.y, ix, iz;
			cVec.x = cx + 0.5f;
			cVec.y = cz + 0.5f;
			MathUtil.chunkCoordsToReal(cVec, rVec);
			
			float xx, zz, distSq;
			xx = rVec.x - camX;
			zz = Math.min(zz = Math.abs(rVec.y - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz);
			distSq = xx * xx + zz * zz;
			
			if (distSq > viewDistance * viewDistance) {
				Chunk chunk = getChunk(cx, cz);
				chunk.onRemove();
				chunks.remove((long) cx << 32 | cz);
				it.remove();
			} else {
				for (int i = 0; chunkToLoad == null && i < 4; i++) {
					ix = cx + (2 * i + 1) % 3 - 1;
					iz = cz + (2 * i + 1) / 3 - 1;
					iz = MathUtil.fitChunkZ(iz);
					p.x = ix;
					p.y = iz;
					
					if (!loadedChunks.contains(p)) {
						cVec.x = ix + 0.5f;
						cVec.y = iz + 0.5f;
						MathUtil.chunkCoordsToReal(cVec, rVec);
						xx = rVec.x - camX;
						zz = Math.min(zz = Math.abs(rVec.y - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz);
						distSq = xx * xx + zz * zz;
						
						if (distSq <= viewDistance * viewDistance) {
							chunkToLoad = p;
						}
					}
				}
			}
		}
		
		if (chunkToLoad != null) {
			setNewChunk(chunkToLoad.x, chunkToLoad.y);
			loadedChunks.add(chunkToLoad);
		}
	}
	
	private void updateHoverHexagon() {
		Vector3f mouseEucl = Main.mousePicker.getCylinderPosition((float) radius);
		
		if (mouseEucl != null) {
			CylVec3 mouse = new CylVec3();
			Vec3f.euclToCylCoords(mouseEucl, mouse);
			float mx = mouse.x, mr = mouse.r;
			double mrv = mouse.rv;
			double min = 0xffffff, temp;
			Hexagon hex = null;
			Chunk c;
			for (Point p : loadedChunks) {
				c = getChunk(p.x, p.y);
				for (BlockColumn hc : c.blockCols) {
					Block b = hc.getBlock(0);
					Hexagon h;
					if (b == null || (h = b.up) == null) continue;
					temp = h.distSq(mx, mrv);
					if (temp < min) {
						min = temp;
						hex = h;
					}
				}
			}
			if (hex != null) {
				if (hoverHexagon != hex) {
					if (hoverHexagon != null) hoverHexagon.deSelect();
					hex.select();
					hoverHexagon = hex;
				}
			} else if (hoverHexagon != null) hoverHexagon.deSelect();
		} else if (hoverHexagon != null) {
			hoverHexagon.deSelect();
			hoverHexagon = null;
		}
		
		if (Mouse.isButtonDown(0)) {
			if (hoverHexagon != null) hoverHexagon.deSelect();
		}
	}
	
	public void render(Camera camera) {
		Shader.hexagon.enable();
		
		BlockRenderer.hexagon.begin();
		BlockRenderer.blockSide.begin();
		for (Point c : loadedChunks) {
			for (BlockColumn col : getChunk(c.x, c.y).blockCols) {
				for (Block b : col.blocks.values()) {
					if (b != null && b.up != hoverHexagon) b.submit();
				}
			}
		}
		BlockRenderer.hexagon.end();
		BlockRenderer.hexagon.render();
		BlockRenderer.blockSide.end();
		BlockRenderer.blockSide.render();
		
		if (hoverHexagon != null) {
			Shader.hexagon.setUniform1f("glow", 1.5f);
			BlockRenderer.hexagon.begin();
			BlockRenderer.blockSide.begin();
			hoverHexagon.block.submit();
			BlockRenderer.hexagon.end();
			BlockRenderer.hexagon.render();
			BlockRenderer.blockSide.end();
			BlockRenderer.blockSide.render();
			Shader.hexagon.setUniform1f("glow", 1f);
		}
	}
	
	public Block getBlock(int x, int y, int z) {
		if (z < 0 || z >= Chunk.chunkSize * Chunk.ringSize) return null;
		int cx = Math.floorDiv(x, Chunk.chunkSize), cz = z / Chunk.chunkSize;
		int mx = x % Chunk.chunkSize, mz = z % Chunk.chunkSize;
		
		Chunk c = getChunk(cx, cz);
		return c.getBlock(mx, y, mz);
	}
	
	private Chunk getChunk(int x, int z) {
		z = MathUtil.fitChunkZ(z);
		
		long index = (long) x << 32 | z;
		Chunk chunk = chunks.get(index);
		if (chunk == null) chunks.put(index, chunk = new Chunk(x, z, tempBlockGenerator.apply(x, z), this));
		
		return chunk;
	}
	
	private Chunk setNewChunk(int x, int z) {
		z = MathUtil.fitChunkZ(z);
		
		long index = (long) x << 32 | z;
		Chunk chunk = chunks.get(index);
		if (chunk != null) chunk.onRemove();
		chunks.put(index, chunk = new Chunk(x, z, tempBlockGenerator.apply(x, z), this));
		
		return chunk;
	}
	
}
